
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// kdtree.cpp*
#include "pbrt.h"
#include "primitive.h"
// KdAccelNode Declarations
struct MailboxPrim {
	MailboxPrim(const Reference<Primitive> &p) {
		    primitive = p;
		    lastMailboxId = -1;
	}
	Reference<Primitive> primitive;
	int lastMailboxId; // identifies last ray that was tested against this mailbox
};


class UnfinishedNodeData {
public:
    // Public Methods
    UnfinishedNodeData() : primNums(NULL) {}

    UnfinishedNodeData(u_int maxNumPrims) : primNums(new u_int[maxNumPrims]) {}

    ~UnfinishedNodeData() {
        if (primNums)
            delete primNums;
    }

    void Print(u_int id, bool showContents = false) {
        printf("Node Data [%d] (unfinished)\n\tnPrims:%d\n\tdepth:%d\n\tbadRefines:%d\n\tweight:%d\n\tbounds: min(%8.2f,%8.2f,%8.2f) max(%8.2f,%8.2f,%8.2f)\n",
               id, nPrims, depth, badRefines, weight, nodeBounds.pMin.x, nodeBounds.pMin.y, nodeBounds.pMin.z, nodeBounds.pMax.x, nodeBounds.pMax.y, nodeBounds.pMax.z);
        if (showContents) {
            printf("\tprimNums");
            for (u_int i = 0; i < nPrims; ++i)
                printf(" [%d]", primNums[i]);
            printf("\n");
        }
    }

    // Public Members
    u_int *primNums; // array containing index (in mailboxPrims) for each prim in this node
    u_int nPrims; // number primitives in primNums
    BBox nodeBounds; // bounds of this node
    int depth; // number of splits we can do from here
    int badRefines; // number of consecutive refines that scored (slightly) worse than parent
    int weight; // sum weight of contained primitives
};


static StatsCounter lazyNodes("Kd-Tree Accelerator", "Lazy kd-tree nodes remaining"); // NOBOOK
static StatsCounter nodesTotal("Kd-Tree Accelerator", "Total kd-tree nodes"); // NOBOOK
struct KdAccelNode {

	// KdAccelNode Methods
	void initLeaf(u_int *primNums, // array of prims inside leaf, each entry contains index of prim in mailboxPrims
            u_int np, // num primitives in leaf
			MailboxPrim *mailboxPrims, // array of mailboxes for all primitives
            MemoryArena &arena) // only passed in if >1 overlapping prim & we need to allocate memory
    {
		// Update kd leaf node allocation statistics
        --lazyNodes;

		static StatsCounter numLeafMade("Kd-Tree Accelerator",
		                                "Leaf kd-tree nodes made");
		++numLeafMade;

		//static StatsCounter maxDepth("Kd-Tree Accelerator",
		//                             "Maximum kd-tree depth");
		//maxDepth.Max(depth);

		static StatsCounter maxLeafPrims("Kd-Tree Accelerator",
			"Maximum number of primitives in leaf node");
		maxLeafPrims.Max(np);


		static StatsRatio leafPrims("Kd-Tree Accelerator",
			"Avg. number of primitives in leaf nodes");
		leafPrims.Add(np, 1);

        // ok, done with stats

        UnfinishedNodeData *dataTmp = data; // primNums actually lives in here, don't delete it

		nPrims = np << 2;
		flags |= 3;

		// Store _MailboxPrim *_s for leaf node
		if (np == 0)
			onePrimitive = NULL;
		else if (np == 1)
			onePrimitive = &mailboxPrims[primNums[0]];
		else {
			primitives = (MailboxPrim **)arena.Alloc(np * sizeof(MailboxPrim *));
			for (u_int i = 0; i < np; ++i)
				primitives[i] = &mailboxPrims[primNums[i]];
		}

        // delete unfinished data
        delete dataTmp;
	}

    void initInterior(int axis, float s) {
		static StatsCounter nodesMade("Kd-Tree Accelerator", "Interior kd-tree nodes made"); // NOBOOK
		++nodesMade; // NOBOOK
        --lazyNodes;

        split = s;
		flags &= ~3;
		flags |= axis;
	}

    void initUnfinished(UnfinishedNodeData *d) {
        ++lazyNodes;
        ++nodesTotal;
        flags = 0xFFFFFFFF;
        data = d;
    }

	float SplitPos() const { return split; }
	int nPrimitives() const { return nPrims >> 2; }
	int SplitAxis() const { return flags & 3; }
    bool IsUnfinished() const { return flags == 0xFFFFFFFF; }
	bool IsLeaf() const { return (flags & 3) == 3; }
	union {
        // Flags: used in All.
        // if Unfinished: 0xFFFFFFFF
        // if Leaf:       last two bits are 11
        // if Interior:   last two bits index x(0), y(1), z(2)
		u_int flags;   
		float split;   // if Interior (no one cares about 2 least significant bits in a float)
		u_int nPrims;  // if Leaf (needs to shift 2 before use)
	};
	union {
		u_int belowChild;           // if Interior (index of below child, belowChild+1 is above child's index)
		MailboxPrim *onePrimitive;  // if Leaf (and <=1 prim overlapping). ptr to 1 mailbox or NULL
		MailboxPrim **primitives;   // if Leaf (and >1 prim overlapping. array of mailbox ptrs
        UnfinishedNodeData *data;   // if Unfinished, ptr to data needed to process
	};
};


struct BoundEdge {

	// BoundEdge Public Methods
	BoundEdge() { }

	BoundEdge(float tt, int pn, bool starting) {
		t = tt;
		primNum = pn;
		type = starting ? START : END;
	}

	bool operator<(const BoundEdge &e) const {
		if (t == e.t)
			return (int)type < (int)e.type;
		else return t < e.t;
	}

	float t;
	int primNum;
	enum { START, END } type;
};


// KdTreeAccel Declarations
struct KdAccelNode;
class  KdTreeAccel : public Aggregate {
public:
	// KdTreeAccel Public Methods
	KdTreeAccel(const vector<Reference<Primitive> > &p, // primitives
		int icost, // intersection cost
        int tcost, // traversal cost
		float ebonus, // empty bonus
        int maxp, int maxDepth);

	BBox WorldBound() const { return bounds; }
	bool CanIntersect() { return true; }
	~KdTreeAccel();
	bool Intersect(const Ray &ray, Intersection *isect);
	bool IntersectP(const Ray &ray);
    u_int Weight() const;

private:
    // KdTreeAccel Private Methods
	void BuildNode(u_int nodeNum);
    void CreateUnfinishedNode(u_int nodeNum, UnfinishedNodeData *data);
    void inline LazyRefinementCheck(Reference<Primitive> &prim);

	// KdTreeAccel Private Data
	int isectCost, traversalCost, maxPrims; // maxPrims is "stop subdividing kd-node if <= maxPrims inside"
	float emptyBonus;

	u_int nMailboxes;
	MailboxPrim *mailboxPrims;
	mutable int curMailboxId; // id to give next ray so it can identify itself in mailbox tests

	KdAccelNode *nodes;
	int nAllocedNodes, nextFreeNode;

	BBox bounds;
	MemoryArena arena;

    // pre-allocated used by each call to BuildNode
	vector<BBox> allPrimBounds; // bounding boxes of all primitives, saved
    BoundEdge *edges[3]; // pre-allocated space for edge info
};

struct KdToDo {
	u_int nodeNum;
	float tmin, tmax;
};


// KdTreeAccel Method Definitions
KdTreeAccel::
    KdTreeAccel(const vector<Reference<Primitive> > &p,
		int icost, int tcost,
		float ebonus, int maxp, int maxDepth)
	: isectCost(icost), traversalCost(tcost),
	maxPrims(maxp), emptyBonus(ebonus)
{
    // the vector of prims we'll be working with in this kd-tree
	vector<Reference<Primitive> > prims;

    // build kd-tree around prims we're given unless we have only 1 primitive AND it is not intersectable
    if (p.size() != 1 || p[0]->CanIntersect())
        prims = p;
    // ... in which case, refine the primitive until this is no longer the case
    else
    {
        for (Reference<Primitive> refineMe = p[0]; !refineMe->CanIntersect(); refineMe = prims[0])
        {
            prims.clear();
            refineMe->Refine(prims);
            if (prims.size() != 1)
                break;
        }
    }

	// Initialize mailboxes for _KdTreeAccel_
	curMailboxId = 0;
	nMailboxes = prims.size();
	mailboxPrims = (MailboxPrim *)AllocAligned(nMailboxes * sizeof(MailboxPrim));
	for (u_int i = 0; i < nMailboxes; ++i)
		new (&mailboxPrims[i]) MailboxPrim(prims[i]);

	// Build kd-tree for accelerator
	nextFreeNode = nAllocedNodes = 0;

	// Compute bounds for kd-tree construction
	allPrimBounds.reserve(prims.size());
	for (u_int i = 0; i < prims.size(); ++i) {
		BBox b = prims[i]->WorldBound();
		allPrimBounds.push_back(b); // save bounds of this prim for later lookup
		bounds = Union(bounds, b); // expand kd-tree bounds
	}

    // edges = {BoundEdge x[2*prims.size()], likewise y, likewise z}
    // 2*prims.size() because we need both starting and ending edge
	for (int i = 0; i < 3; ++i)
		edges[i] = new BoundEdge[2*prims.size()];
    
    // set data for first unfinished node
    UnfinishedNodeData *data = new UnfinishedNodeData(prims.size());
    data->badRefines = 0;
    data->nodeBounds = bounds;
    data->weight = 0;
    for (u_int i = 0; i < prims.size(); ++i)
    {
        data->primNums[i] = i; // all prims overlap initial node
        data->weight += prims[i]->Weight();
    }

    // set maxDepth
	if (maxDepth > 0)
        data->depth = maxDepth;
    else
		data->depth = Round2Int(8 + 1.3f * Log2Int(float(data->weight)));

    data->nPrims = prims.size();

    // set first node
    CreateUnfinishedNode(0, data);
}


KdTreeAccel::~KdTreeAccel() {
    // delete dynamically allocated UnfinishedNodeData in unfinished nodes
    for (int i = 0; i < nextFreeNode; ++i)
        if (nodes[i].IsUnfinished())
            delete nodes[i].data;

	for (u_int i = 0; i < nMailboxes; ++i)
		mailboxPrims[i].~MailboxPrim();
	FreeAligned(mailboxPrims);
	FreeAligned(nodes);
    for (u_int i = 0; i < 3; ++i)
        delete edges[i];
}

void KdTreeAccel::CreateUnfinishedNode(u_int nodeNum, UnfinishedNodeData *data)
{
	Assert(nodeNum == nextFreeNode); // NOBOOK

	// Get next free node from _nodes_ array
    // if all allocated nodes used up, reallocate with twice as 
    // much memory (but at least 512) & copy over old values
	if (nextFreeNode == nAllocedNodes) {
		int nAlloc = max(2 * nAllocedNodes, 512);
		KdAccelNode *n = (KdAccelNode *)AllocAligned(nAlloc *
			sizeof(KdAccelNode));
		if (nAllocedNodes > 0) {
			memcpy(n, nodes,
			       nAllocedNodes * sizeof(KdAccelNode));
			FreeAligned(nodes);
		}
		nodes = n;
		nAllocedNodes = nAlloc;
	}
	++nextFreeNode;

    // set up node
    nodes[nodeNum].initUnfinished(data);
}


// if we can't actually intersect this primitive, replace it (in the mailbox) with a kd-tree which
// contains the primitive's refined elements
void KdTreeAccel::LazyRefinementCheck(Reference<Primitive> &prim) 
{
    if (!prim->CanIntersect())
    {
        vector<Reference<Primitive> > p(1);
        p[0] = prim;
        prim = new KdTreeAccel(p, isectCost, traversalCost, emptyBonus,
                               maxPrims, -1);
    }
}




void KdTreeAccel::BuildNode(u_int nodeNum)
{
    Assert(nodes[nodeNum].IsUnfinished());

    UnfinishedNodeData *data = nodes[nodeNum].data;

	// Initialize leaf node if termination criteria met
	if ((int)data->nPrims <= maxPrims || data->depth == 0) {
		nodes[nodeNum].initLeaf(data->primNums, data->nPrims,
                                mailboxPrims, arena);

		return;
	}

	// Initialize interior node and continue recursion
	// Choose split axis position for interior node
	int bestAxis = -1, bestOffset = -1, bestAboveCount = -1, bestWeightBelow, bestWeightAbove;
	float bestCost = INFINITY;
	float oldCost = isectCost * float(data->weight);

	Vector d = data->nodeBounds.pMax - data->nodeBounds.pMin; // diagonal across bounding box
	float totalSA = (2.f * (d.x*d.y + d.x*d.z + d.y*d.z));
	float invTotalSA = 1.f / totalSA; // to avoid divides later

	// Choose which axis to split along
	int axis;
	if (d.x > d.y && d.x > d.z)
        axis = 0;
	else
        axis = (d.y > d.z) ? 1 : 2;
	int retries = 0;
	retrySplit:

	// Initialize edges for _axis_
	for (u_int i = 0; i < data->nPrims; ++i) {
		int pn = data->primNums[i];
		const BBox &bbox = allPrimBounds[pn];
		edges[axis][2 * i] =
		    BoundEdge(bbox.pMin[axis], pn, true);
		edges[axis][2 * i + 1] =
			BoundEdge(bbox.pMax[axis], pn, false);
	}
	sort(&edges[axis][0], &edges[axis][2 * data->nPrims]);

	// Compute cost of all splits for _axis_ to find best
	int nBelow = 0, nAbove = data->nPrims, weightBelow = 0, weightAbove = data->weight;
	for (u_int i = 0; i < 2 * data->nPrims; ++i)
    {
        int primWeight = mailboxPrims[ edges[axis][i].primNum ].primitive->Weight();

		if (edges[axis][i].type == BoundEdge::END)
        {
            --nAbove;
            weightAbove -= primWeight;
        }
		float edget = edges[axis][i].t;
		if (edget > data->nodeBounds.pMin[axis] &&
			edget < data->nodeBounds.pMax[axis])
        {
			// Compute cost for split at _i_th edge
			int otherAxis[3][2] = { {1,2}, {0,2}, {0,1} };
			int otherAxis0 = otherAxis[axis][0];
			int otherAxis1 = otherAxis[axis][1];

            // probability of ray intersecting A, given intersection of parent, is SA(a) / SA(parent)
			float belowSA = 2 * (d[otherAxis0] * d[otherAxis1]
                                 + (edget - data->nodeBounds.pMin[axis])
                                   * (d[otherAxis0] + d[otherAxis1]));

			float aboveSA = 2 * (d[otherAxis0] * d[otherAxis1]
                                 + (data->nodeBounds.pMax[axis] - edget)
                                   * (d[otherAxis0] + d[otherAxis1]));

			float pBelow = belowSA * invTotalSA;
			float pAbove = aboveSA * invTotalSA;
			float eb = (nAbove == 0 || nBelow == 0) ? emptyBonus : 0.f;
			float cost = traversalCost + isectCost * (1.f - eb) *
				(pBelow * weightBelow + pAbove * weightAbove);

			// Update best split if this is lowest cost so far
			if (cost < bestCost)  {
				bestCost = cost;
				bestAxis = axis;
				bestOffset = i;
                bestAboveCount = nAbove;
                bestWeightBelow = weightBelow;
                bestWeightAbove = weightAbove;
			}
		}
		if (edges[axis][i].type == BoundEdge::START)
        {
            ++nBelow;
            weightBelow += primWeight;
        }
	}
	Assert(nBelow == data->nPrims && nAbove == 0); // NOBOOK

    // if we found no splits, try another axis
    if (bestAxis == -1 && retries < 2) {
		++retries;
		axis = (axis + 1) % 3;
		goto retrySplit; // daaaaaaaaaaaaamn it's a goto
	}

	if (bestCost > oldCost)
        ++data->badRefines;

    // Create leaf if no good splits were found
	if ((bestCost > 4.f * oldCost && data->nPrims < 16) // lots of magic #s in this code
        || bestAxis == -1
        || data->badRefines == 3) // it's magic
    {
		nodes[nodeNum].initLeaf(data->primNums, data->nPrims,
		                     mailboxPrims, arena);
		return;
	}

    // Set up data for two child nodes
    u_int nPrims = data->nPrims;

    UnfinishedNodeData *dataBelow = data, // re-use current node's data as it's below-child's data
                       *dataAbove = new UnfinishedNodeData(bestAboveCount);
    // ...badRefines
    dataAbove->badRefines = dataBelow->badRefines;
    // ...depth
    dataAbove->depth = --dataBelow->depth;
    // ...weight
    dataAbove->weight = bestWeightAbove;
    dataBelow->weight = bestWeightBelow;
    // ...primNums: classify with respect to split
    // ...nPrims
	int n0 = 0; // next available index in dataBelow->primNums (below split)
    u_int *prims0 = dataBelow->primNums;
	for (int i = 0; i < bestOffset; ++i)
		if (edges[bestAxis][i].type == BoundEdge::START)
			prims0[n0++] = edges[bestAxis][i].primNum;
    dataBelow->nPrims = n0;

    int n1 = 0; // next available index in dataAbove->primNums (above split)
    u_int *prims1 = dataAbove->primNums;
	for (u_int i = bestOffset+1; i < (nPrims << 1); ++i)
		if (edges[bestAxis][i].type == BoundEdge::END)
			prims1[n1++] = edges[bestAxis][i].primNum;
    dataAbove->nPrims = n1;

    // ...bounds
	float tsplit = edges[bestAxis][bestOffset].t; // save split
	dataAbove->nodeBounds = dataBelow->nodeBounds; // initialize sub-nodes' bboxes to be parent's bbox
	dataBelow->nodeBounds.pMax[bestAxis] = dataAbove->nodeBounds.pMin[bestAxis] = tsplit; // resize bboxes appropriately

    // Finish node, mark as interior
	nodes[nodeNum].initInterior(bestAxis, tsplit);

    // allocate room for both children. point parent to first child in sequence
    nodes[nodeNum].belowChild = nextFreeNode;
    CreateUnfinishedNode(nextFreeNode, dataBelow);
    CreateUnfinishedNode(nextFreeNode, dataAbove);

}


bool KdTreeAccel::Intersect(const Ray &ray,
		Intersection *isect)
{
	// Compute initial parametric range of ray inside kd-tree extent
	float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
		return false;

	// Prepare to traverse kd-tree for ray
	int rayId = curMailboxId++;
	Vector invDir(1.f/ray.d.x, 1.f/ray.d.y, 1.f/ray.d.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;

	// Traverse kd-tree nodes in order for ray
	bool hit = false;
    u_int nodeNum = 0;
	const KdAccelNode *node = &nodes[nodeNum];
	while (node != NULL)
    {
		// Bail out if we found a hit closer than the current node
		if (ray.maxt < tmin) // meaning we're looking at the todo stack though we just found a collision (this should be a while-loop condition))
            break;

        // If node is unfinished, build it
        if (node->IsUnfinished()) {
            BuildNode(nodeNum);
            node = &nodes[nodeNum]; // set here, since BuildNode may have rebuilt array
        }

        //static StatsCounter nodesTraversed("Kd-Tree Accelerator", //NOBOOK
		//	"Number of kd-tree nodes traversed by normal rays"); //NOBOOK
		//++nodesTraversed; //NOBOOK
		if (!node->IsLeaf()) // if internal
        {
			// Process kd-tree interior node
			// Compute parametric distance along ray to split plane
			int axis = node->SplitAxis();
			float tplane = (node->SplitPos() - ray.o[axis]) // distance to plane
                           * invDir[axis];

            // Get node children pointers for ray
			u_int firstChildNum, secondChildNum;
			int belowFirst = ray.o[axis] <= node->SplitPos();
			if (belowFirst) {
				firstChildNum = node->belowChild;
				secondChildNum = node->belowChild + 1; // aboveChild
			}
			else {
				firstChildNum = node->belowChild + 1; // aboveChild
				secondChildNum = node->belowChild;
			}

			// Advance to next child node, possibly enqueue other child.

            // if ray crosses split-plane once outside bbox, we never need consider secondChild
            // if ray crosses split-plane by going in reverse, we never need consider secondChild
			if (tplane > tmax || tplane < 0) 
				nodeNum = firstChildNum;

            // if we cross plane before we even get to bbox, we never need consider firstChild
			else if (tplane < tmin)
				nodeNum = secondChildNum;

            // else ray goes through both, process firstChild 1st, push secondChild onto todo stack
			else
            {
				// Enqueue _secondChild_ in todo list
				todo[todoPos].nodeNum = secondChildNum;
				todo[todoPos].tmin = tplane; // where we entered this child node
				todo[todoPos].tmax = tmax; // where we shot out of our bbox
				++todoPos;
				nodeNum = firstChildNum; // process firstChild
				tmax = tplane;
			}
		}
		else
        {
			// Check for intersections inside leaf node
			u_int nPrimitives = node->nPrimitives();
			if (nPrimitives == 1)
            {
				MailboxPrim *mp = node->onePrimitive;
				// Check one primitive inside leaf node
				if (mp->lastMailboxId != rayId)
                {
                    LazyRefinementCheck(mp->primitive);
					mp->lastMailboxId = rayId;
				    if (mp->primitive->Intersect(ray, isect))
					    hit = true;
				}
			}
			else
            {
				MailboxPrim **prims = node->primitives;
				for (u_int i = 0; i < nPrimitives; ++i)
                {
					MailboxPrim *mp = prims[i];
					// Check one primitive inside leaf node
					if (mp->lastMailboxId != rayId)
                    {
                        LazyRefinementCheck(mp->primitive);
						mp->lastMailboxId = rayId;
						if (mp->primitive->Intersect(ray, isect))
							hit = true;
					}
				}
			}
			// Grab next node to process from todo list
			if (todoPos > 0)
            {
				--todoPos;
				nodeNum = todo[todoPos].nodeNum;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
			}
			else
				break;
		}

        node = &nodes[nodeNum];
	}
	return hit;
}


bool KdTreeAccel::IntersectP(const Ray &ray)
{
	// Compute initial parametric range of ray inside kd-tree extent
	float tmin, tmax;
	if (!bounds.IntersectP(ray, &tmin, &tmax))
		return false;

    // Prepare to traverse kd-tree for ray
	int rayId = curMailboxId++;
	Vector invDir(1.f/ray.d.x, 1.f/ray.d.y, 1.f/ray.d.z);
	#define MAX_TODO 64
	KdToDo todo[MAX_TODO];
	int todoPos = 0;

    u_int nodeNum = 0;
    const KdAccelNode *node = &nodes[nodeNum];
	while (node != NULL)
    {
        // If node is unfinished, build it
        if (node->IsUnfinished())
        {
            BuildNode(nodeNum);
            node = &nodes[nodeNum]; // set here, since BuildNode may have rebuilt array
        }

		// Update kd-tree shadow ray traversal statistics
		//static StatsCounter nodesTraversed("Kd-Tree Accelerator",
		//	"Number of kd-tree nodes traversed by shadow rays");
		//++nodesTraversed;
		if (node->IsLeaf())
        {
			// Check for shadow ray intersections inside leaf node
			u_int nPrimitives = node->nPrimitives();
			if (nPrimitives == 1)
            {
				MailboxPrim *mp = node->onePrimitive;
				if (mp->lastMailboxId != rayId)
                {
                    LazyRefinementCheck(mp->primitive);
					mp->lastMailboxId = rayId;
					if (mp->primitive->IntersectP(ray))
						return true;
				}
			}
			else
            {
				MailboxPrim **prims = node->primitives;
				for (u_int i = 0; i < nPrimitives; ++i)
                {
					MailboxPrim *mp = prims[i];
					if (mp->lastMailboxId != rayId)
                    {
                        LazyRefinementCheck(mp->primitive);
						mp->lastMailboxId = rayId;
						if (mp->primitive->IntersectP(ray))
							return true;
					}
				}
			}

			// Grab next node to process from todo list
			if (todoPos > 0)
            {
				--todoPos;
				nodeNum = todo[todoPos].nodeNum;
				tmin = todo[todoPos].tmin;
				tmax = todo[todoPos].tmax;
			}
			else
				break;
		}
		else
        {
			// Process kd-tree interior node
			// Compute parametric distance along ray to split plane
			int axis = node->SplitAxis();
			float tplane = (node->SplitPos() - ray.o[axis]) *
				invDir[axis];

            // Get node children pointers for ray
			u_int firstChildNum, secondChildNum;
			int belowFirst = ray.o[axis] <= node->SplitPos();
			if (belowFirst)
            {
				firstChildNum = node->belowChild;
				secondChildNum = node->belowChild + 1; // aboveChild
			}
			else
            {
				firstChildNum = node->belowChild + 1; // aboveChild
				secondChildNum = node->belowChild;
			}

            // Advance to next child node, possibly enqueue other child
			if (tplane > tmax || tplane < 0)
				nodeNum = firstChildNum;
			else if (tplane < tmin)
				nodeNum = secondChildNum;
			else
            {
				// Enqueue _secondChild_ in todo list
				todo[todoPos].nodeNum = secondChildNum;
				todo[todoPos].tmin = tplane;
				todo[todoPos].tmax = tmax;
				++todoPos;
				nodeNum = firstChildNum;
				tmax = tplane;
			}
		}

        node = &nodes[nodeNum];
	}
	return false;
}

// weight of tree is combined weight of primitives in mailboxes
u_int KdTreeAccel::Weight() const {
    u_int t = 0;
    for (u_int i = 0; i < nMailboxes; ++i)
        t += mailboxPrims[i].primitive->Weight();
    return t;
}


extern "C" DLLEXPORT Primitive *CreateAccelerator(const vector<Reference<Primitive> > &prims,
		const ParamSet &ps) {
	int isectCost = ps.FindOneInt("intersectcost", 80);
	int travCost = ps.FindOneInt("traversalcost", 1);
	float emptyBonus = ps.FindOneFloat("emptybonus", 0.5f);
	int maxPrims = ps.FindOneInt("maxprims", 1);
	int maxDepth = ps.FindOneInt("maxdepth", -1);

    //ProgressReporter progress(1, "Building KDTree");
    KdTreeAccel* accel = new KdTreeAccel(prims, isectCost, travCost,
		                               emptyBonus, maxPrims, maxDepth);
    //progress.Update();
    //progress.Done();

    return accel;
}
